/*
 * (c) Copyright 2010 by Volker Bergmann. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted under the terms of the
 * GNU General Public License (GPL).
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * WITHOUT A WARRANTY OF ANY KIND. ALL EXPRESS OR IMPLIED CONDITIONS,
 * REPRESENTATIONS AND WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE
 * HEREBY EXCLUDED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.databene.benerator.nullable;

import java.util.Random;

import org.databene.benerator.GeneratorContext;
import org.databene.benerator.InvalidGeneratorSetupException;
import org.databene.benerator.wrapper.ProductWrapper;

/**
 * Proxy implementation of the {@link NullableGenerator} interface.<br/><br/>
 * Created: 26.01.2010 10:32:58
 * @since 0.6.0
 * @author Volker Bergmann
 */
public class NullInjectingGeneratorProxy<E> extends NullableGeneratorWrapper<E, E> {
	
	private float nullQuota;
	
	private Random random;
	boolean closed;

    public NullInjectingGeneratorProxy(NullableGenerator<E> source, double nullQuota) {
    	super(source);
    	if (nullQuota < 0 || nullQuota > 1)
    		throw new IllegalArgumentException("Illegal null quota: " + nullQuota);
	    this.nullQuota = (float) nullQuota;
		random = new Random();
		closed = false;
    }

    public Class<E> getGeneratedType() {
    	return source.getGeneratedType();
    }

	@Override
    public void init(GeneratorContext context) throws InvalidGeneratorSetupException {
		source.init(context);
    	super.init(context);
    }

    /**
     * First checks if a null value should be returned, if so it returns a wrapper that wraps a null value, 
     * otherwise a wrapper that wraps a value generated by the underlying source generator. If that generator
     * is not available any more, <code>null</code> is returned. 
     */
    public ProductWrapper<E> generate(ProductWrapper<E> wrapper) {
    	if (closed)
    		return null;
    	if (random.nextFloat() < nullQuota) {
    		wrapper.product = null;
    		return wrapper;
    	}
    	return source.generate(wrapper);
    }

	@Override
	public String toString() {
	    return getClass().getSimpleName() + "[" + nullQuota + ", " + source + "]";
	}
    
}
